"""BI CLI."""

import functools
from enum import Enum
from typing import Annotated, Optional

import typer
import yaml
from pydantic import RootModel
from rich.table import Table
from slices_bi_client.exceptions import ApiException
from slices_bi_client.models.disk_image import DiskImage
from slices_bi_client.models.disk_image_create import DiskImageCreate
from urllib3.connection import HTTPSConnection
from urllib3.exceptions import MaxRetryError, NewConnectionError

from . import console, err_console
from .bi_context import BIContext
from .utils import friendly_datetime, get_error_detail

disk_images_app = typer.Typer(no_args_is_help=True)


def handle_some_exceptions(func):
    """Handle specific exceptions by giving the user a useful error message instead of a confusing stacktrace."""

    @functools.wraps(func)
    def wrapper_handle_some_exceptions(*args, **kwargs):
        """Function wrapper for handle_some_exceptions."""  # noqa: D401
        try:
            return func(*args, **kwargs)
        except ApiException as e:
            if e.status in (
                502,
                503,
                504,
            ):
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise
        except MaxRetryError as e:
            if isinstance(e.reason, NewConnectionError | HTTPSConnection):
                # example:
                #   NewConnectionError -> [Failed to establish a new connection: Errno 113] No route to host
                #   HTTPSConnection -> Failed to establish a new connection: [Errno 111] Connection refused
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e.reason}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise

    return wrapper_handle_some_exceptions


@disk_images_app.command(name="register", rich_help_panel="Admin Commands")
@handle_some_exceptions
def register_image(
    ctx: BIContext,
    friendly_name: Annotated[str, typer.Option(prompt=True, show_default=False)],
    location: Annotated[str, typer.Option(prompt=True, show_default=False)],
    default_username: Annotated[str, typer.Option(prompt=True, show_default=False)],
    min_disk: Annotated[
        Optional[int],
        typer.Option(show_default=False, help="The minimum size of the root disk, in MB."),
    ] = None,
    min_ram: Annotated[
        Optional[int],
        typer.Option(show_default=False, help="The minimum required amount of RAM for this image, MiB."),
    ] = None,
    tags: Annotated[Optional[list[str]], typer.Option(show_default=False)] = None,
):
    """Register disk images.

    Only available to admin users.
    """

    dic = DiskImageCreate(
        friendly_name=friendly_name,
        cluster_id=ctx.obj.cluster_id,
        location=location,
        default_username=default_username,
    )
    if min_disk:
        dic.min_disk_mb = min_disk
    if min_ram:
        dic.min_ram_mib = min_ram
    if tags:
        dic.tags = tags

    try:
        disk_image = ctx.obj.api.get_bi_images_api(ctx.obj.site_id, ctx.obj.require_project_id()).register_disk_image(
            dic
        )
        console.print(f"[bold green]:cd: Registered {friendly_name} as {disk_image.id} :white_check_mark:")
    except ApiException as err:
        err_console.print("[bold red]Could not register disk image[/] :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(code=1) from None


class ListFormat(Enum):
    """List Formats."""

    table = "table"
    json = "json"
    yaml = "yaml"


DiskImages = RootModel[list[DiskImage]]


@disk_images_app.command(name="list", rich_help_panel="Commands")
@handle_some_exceptions
def list_images(
    ctx: BIContext,
    format: Annotated[
        ListFormat,
        typer.Option("--format", "-f", help="Format in which the images should be listed."),
    ] = "table",  # type: ignore
):
    """List disk images."""
    images_page = ctx.obj.api.get_bi_images_api(ctx.obj.site_id, ctx.obj.require_project_id()).get_disk_images()

    if format == ListFormat.table:
        table = Table("ID", "Name", "Created At", "Tags", title="Images")
        for disk_image in images_page.disk_images:
            table.add_row(
                disk_image.id,
                disk_image.friendly_name,
                friendly_datetime(disk_image.created_at) if disk_image.created_at else "-",
            )

        console.print(table)

    elif format == ListFormat.json:
        console.print_json(DiskImages(images_page.disk_images).model_dump_json(by_alias=True, exclude_unset=True))
    elif format == ListFormat.yaml:
        console.print(
            yaml.safe_dump(
                DiskImages(images_page.disk_images).model_dump(mode="json", by_alias=True, exclude_unset=True)
            )
        )
